@@ -50,6 +50,7 @@ x_init()
 	if (setsig(&sigtraps[SIGWINCH], x_sigwinch, SS_RESTORE_ORIG|SS_SHTRAP))
 		sigtraps[SIGWINCH].flags |= TF_SHELL_USES;
 # endif /* SIGWINCH */
+	got_sigwinch = 1; /* force initial check */
 	check_sigwinch();
 #endif /* TIOCGWINSZ */
 
@@ -306,7 +307,6 @@ x_mode(onoff)
  * RETURN VALUE:
  *      length
  */
- 
 int
 promptlen(cp, spp)
     const char  *cp;
@@ -314,37 +314,35 @@ promptlen(cp, spp)
 {
     int count = 0;
     const char *sp = cp;
+    char delimiter = 0;
+    int indelimit = 0;
 
-    while (*cp) {
-	if (*cp == '\n' || *cp == '\r') {
+    /* Undocumented AT&T ksh feature:
+     * If the second char in the prompt string is \r then the first char
+     * is taken to be a non-printing delimiter and any chars between two
+     * instances of the delimiter are not considered to be part of the
+     * prompt length
+     */
+    if (*cp && cp[1] == '\r') {
+	delimiter = *cp;
+	cp += 2;
+    }
+
+    for (; *cp; cp++) {
+	if (indelimit && *cp != delimiter)
+	    ;
+	else if (*cp == '\n' || *cp == '\r') {
 	    count = 0;
-	    cp++;
-	    sp = cp;
+	    sp = cp + 1;
 	} else if (*cp == '\t') {
 	    count = (count | 7) + 1;
-	    cp++;
 	} else if (*cp == '\b') {
 	    if (count > 0)
 		count--;
-	    cp++;
-	}
-#if 1
+	} else if (*cp == delimiter)
+	    indelimit = !indelimit;
 	else
-	  cp++, count++;
-#else
-	else if (*cp++ != '!')
-	  count++;
-	else if (*cp == '!') {
-	    cp++;
 	    count++;
-	} else {
-	    register int i = source->line + 1;
-
-	    do
-		count++;
-	    while ((i /= 10) > 0);
-	}
-#endif /* 1 */
     }
     if (spp)
 	*spp = sp;
@@ -820,6 +818,7 @@ add_glob(str, slen)
 {
 	char *toglob;
 	char *s;
+	bool_t saw_slash = FALSE;
 
 	if (slen < 0)
 		return (char *) 0;
@@ -829,8 +828,9 @@ add_glob(str, slen)
 
 	/*
 	 * If the pathname contains a wildcard (an unquoted '*',
-	 * '?', or '[') or parameter expansion ('$'), then it is globbed
-	 * based on that value (i.e., without the appended '*').
+	 * '?', or '[') or parameter expansion ('$'), or a ~username
+	 * with no trailing slash, then it is globbed based on that
+	 * value (i.e., without the appended '*').
 	 */
 	for (s = toglob; *s; s++) {
 		if (*s == '\\' && s[1])
@@ -838,8 +838,10 @@ add_glob(str, slen)
 		else if (*s == '*' || *s == '[' || *s == '?' || *s == '$'
 			 || (s[1] == '(' /*)*/ && strchr("*+?@!", *s)))
 			break;
+		else if (ISDIRSEP(*s))
+			saw_slash = TRUE;
 	}
-	if (!*s) {
+	if (!*s && (*toglob != '~' || saw_slash)) {
 		toglob[slen] = '*';
 		toglob[slen + 1] = '\0';
 	}
