@@ -1,3 +1,9 @@
+#ifndef KSH_TIMES_H
+# define KSH_TIMES_H
+
+/* Needed for clock_t on some systems (ie, NeXT in non-posix mode) */
+#include "ksh_time.h"
+
 #include <sys/times.h>
 
 #ifdef TIMES_BROKEN
@@ -9,3 +15,4 @@ extern clock_t	ksh_times ARGS((struct tms *));
 #ifdef HAVE_TIMES
 extern clock_t	times ARGS((struct tms *));
 #endif /* HAVE_TIMES */
+#endif /* KSH_TIMES_H */
