@@ -3,6 +3,9 @@
  * John R. MacMillan
  */
 
+#include "config.h"
+
+#ifdef KSH
 #include "sh.h"
 #include "ksh_stat.h"
 #include "ksh_time.h"
@@ -16,25 +19,19 @@ typedef struct mbox {
 	time_t		mb_mtime;	/* mtime of mail file */
 } mbox_t;
 
-struct mailmsg {
-	char		*msg;		/* Text of message */
-	struct mailmsg	*next;		/* Next message */
-};
-
 /*
  * $MAILPATH is a linked list of mboxes.  $MAIL is a treated as a
  * special case of $MAILPATH, where the list has only one node.  The
  * same list is used for both since they are exclusive.
  */
 
-static mbox_t  *mplist = NULL;
-static mbox_t  mbox = { NULL, NULL, NULL, 0 };
-static time_t	mlastchkd = 0;	/* when mail was last checked */
-static struct mailmsg *mmsgs = NULL;	/* Messages to be printed */
+static mbox_t  *mplist;
+static mbox_t  mbox;
+static time_t	mlastchkd;	/* when mail was last checked */
 
 static void     munset      ARGS((mbox_t *mlist)); /* free mlist and mval */
 static mbox_t * mballoc     ARGS((char *p, char *m)); /* allocate a new mbox */
-static void     maddmsg     ARGS((mbox_t *mbp));
+static void     mprintit    ARGS((mbox_t *mbp));
 
 void
 mcheck()
@@ -69,7 +66,7 @@ mcheck()
 				if (stbuf.st_size
 				    && mbp->mb_mtime != stbuf.st_mtime
 				    && stbuf.st_atime <= stbuf.st_mtime)
-					maddmsg(mbp);
+					mprintit(mbp);
 				mbp->mb_mtime = stbuf.st_mtime;
 			} else {
 				/*
@@ -111,7 +108,7 @@ mpset(mptoparse)
 
 	munset( mplist );
 	mplist = NULL;
-	mval = strsave(mptoparse, APERM);
+	mval = str_save(mptoparse, APERM);
 	while (mval) {
 		mpath = mval;
 		if ((mval = strchr(mval, PATHSEP)) != NULL) {
@@ -175,35 +172,16 @@ mballoc(p, m)
 	return(mbp);
 }
 
-void
-mprint()
-{
-	struct mailmsg *mm;
-
-	while ((mm = mmsgs) != NULL) {
-		shellf("%s\n", mm->msg);
-		afree((void *)mm->msg, APERM);
-		mmsgs = mm->next;
-		afree((void *)mm, APERM);
-	}
-}
-
 static void
-maddmsg( mbp )
+mprintit( mbp )
 mbox_t	*mbp;
 {
-	struct mailmsg	*message;
 	struct tbl	*vp;
 
-	message = (struct mailmsg *)alloc(sizeof(struct mailmsg), APERM);
-	setstr((vp = typeset("_", LOCAL, 0, 0, 0)), mbp->mb_path);
+	setstr((vp = local("_", FALSE)), mbp->mb_path);
 
-	if (mbp->mb_msg)
-		message->msg = strsave(substitute(mbp->mb_msg,0),APERM);
-	else
-		message->msg = strsave(substitute(MBMESSAGE,0),APERM);
+	shellf("%s\n", substitute(mbp->mb_msg ? mbp->mb_msg : MBMESSAGE, 0));
 
 	unset(vp, 0);
-	message->next = mmsgs;
-	mmsgs = message;
 }
+#endif /* KSH */
