@@ -680,6 +680,14 @@ EXTERN	int	x_cols I__(80);	/* tty columns */
 #define OBRACE	'{'
 #define CBRACE	'}'
 
+/* Determine the location of the system (common) profile */
+#ifndef KSH_SYSTEM_PROFILE
+# ifdef __NeXT
+#  define KSH_SYSTEM_PROFILE "/etc/profile.std"
+# else /* __NeXT */
+#  define KSH_SYSTEM_PROFILE "/etc/profile"
+# endif /* __NeXT */
+#endif /* KSH_SYSTEM_PROFILE */
 
 #include "shf.h"
 #include "table.h"
