@@ -34,6 +34,7 @@ static void	syntaxerr	ARGS((const char *what))
 static void	multiline_push ARGS((struct multiline_state *save, int tok));
 static void	multiline_pop ARGS((struct multiline_state *saved));
 static int	assign_command ARGS((char *s));
+static int	inalias ARGS((struct source *s));
 #ifdef KSH
 static int	dbtestp_isa ARGS((Test_env *te, Test_meta meta));
 static const char *dbtestp_getopnd ARGS((Test_env *te, Test_op op,
@@ -120,8 +121,11 @@ c_list()
 
 	t = andor();
 	if (t != NULL) {
+		/* Token has always been read/rejected at this point, so
+		 * we don't worray about what flags to pass token()
+		 */
 		while ((c = token(0)) == ';' || c == '&' || c == COPROC ||
-		       (c == '\n' && (multiline.on || source->type == SALIAS)))
+		       (c == '\n' && (multiline.on || inalias(source))))
 		{
 			if (c == '&' || c == COPROC) {
 				int type = c == '&' ? TASYNC : TCOPROC;
@@ -205,7 +209,10 @@ get_command(cf)
 	XPinit(args, 16);
 	XPinit(vars, 16);
 
-	if (multiline.on)
+	/* Don't want to pass CONTIN if reading interactively as just hitting
+	 * return would print PS2 instead of PS1.
+	 */
+	if (multiline.on || inalias(source))
 		cf = CONTIN;
 	syniocf = KEYWORD|ALIAS;
 	switch (c = token(cf|KEYWORD|ALIAS|VARASN)) {
@@ -282,6 +289,7 @@ get_command(cf)
 		t = nested(TBRACE, '{', '}');
 		break;
 
+#ifdef KSH
 	  case MDPAREN:
 	  {
 		static const char let_cmd[] = { CHAR, 'l', CHAR, 'e',
@@ -294,6 +302,7 @@ get_command(cf)
 		XPput(args, yylval.cp);
 		break;
 	  }
+#endif /* KSH */
 
 #ifdef KSH
 	  case DBRACKET: /* [[ .. ]] */
@@ -656,8 +665,8 @@ const	struct tokeninfo {
 	{ "&&",		LOGAND,	FALSE },
 	{ "||",		LOGOR,	FALSE },
 	{ ";;",		BREAK,	FALSE },
-	{ "((",		MDPAREN, FALSE },
 #ifdef KSH
+	{ "((",		MDPAREN, FALSE },
 	{ "|&",		COPROC,	FALSE },
 #endif /* KSH */
 	/* and some special cases... */
@@ -808,6 +817,17 @@ assign_command(s)
 		|| (c == 't' && strcmp(s, "typeset") == 0);
 }
 
+/* Check if we are in the middle of reading an alias */
+static int
+inalias(s)
+	struct source *s;
+{
+	for (; s && s->type == SALIAS; s = s->next)
+		if (!(s->flags & SF_ALIASEND))
+			return 1;
+	return 0;
+}
+
 
 #ifdef KSH
 /* Order important - indexed by Test_meta values
